import hashlib
from enum import StrEnum
from io import TextIOWrapper
from typing import Any, TextIO

from pydantic import BaseModel, ConfigDict

from labels.model.package_manager import PackageManager


class DependencyType(StrEnum):
    DIRECT = "DIRECT"
    TRANSITIVE = "TRANSITIVE"
    ROOT = "ROOT"
    UNDETERMINABLE = "UNDETERMINABLE"


class Scope(StrEnum):
    BUILD = "BUILD"
    RUN = "RUN"
    UNDETERMINABLE = "UNDETERMINABLE"


class Coordinates(BaseModel):
    real_path: str
    file_system_id: str | None = None
    line: int | None = None


class DependencyChain(BaseModel):
    depth: int
    chain: list[str]


class Location(BaseModel):
    scope: Scope = Scope.UNDETERMINABLE
    coordinates: Coordinates | None = None
    access_path: str | None = None
    dependency_type: DependencyType = DependencyType.UNDETERMINABLE
    reachable_cves: list[str] = []
    top_parents: list[str] | None = None
    dependency_chains: list[DependencyChain] | None = None
    package_manager: PackageManager = PackageManager.UNKNOWN

    def path(self) -> str:
        path = self.access_path or (self.coordinates.real_path if self.coordinates else None)
        if not path:
            error_msg = "Both access_path and coordinates.real_path are empty"
            raise ValueError(error_msg)
        return path

    def location_id(self) -> str:
        path = self.path()
        line = self.coordinates.line if self.coordinates else 0
        location_str = f"{path}:{line}"
        return hashlib.sha256(location_str.encode()).hexdigest()[:16]

    def output_dependency_type(self) -> DependencyType:
        """Return the appropriate dependency_type for public SBOM output.

        ROOT is an internal type used for calculations like top_parents.
        In the output it is represented as DIRECT to avoid exposing implementation details.
        """
        return (
            DependencyType.DIRECT
            if self.dependency_type == DependencyType.ROOT
            else self.dependency_type
        )

    def serialize(self) -> dict[str, Any]:  # type: ignore[explicit-any]
        return {
            "path": self.path(),
            "line": self.coordinates.line if self.coordinates and self.coordinates.line else None,
            "layer": self.coordinates.file_system_id
            if self.coordinates and self.coordinates.file_system_id
            else None,
            "dependency_type": self.output_dependency_type(),
            "scope": self.scope.value,
            "reachable_cves": self.reachable_cves,
            "location_id": self.location_id(),
            "top_parents": self.top_parents,
            "dependency_chains": [dc.model_dump() for dc in self.dependency_chains]  # type: ignore[misc]
            if self.dependency_chains is not None
            else None,
            "package_manager": self.package_manager.value,
        }


class LocationReadCloser(BaseModel):
    location: Location
    read_closer: TextIO | TextIOWrapper
    model_config = ConfigDict(arbitrary_types_allowed=True)



from enum import Enum

from pydantic import BaseModel, ConfigDict


class RelationshipType(Enum):
    OWNERSHIP_BY_FILE_OVERLAP_RELATIONSHIP = "ownership-by-file-overlap"
    EVIDENT_BY_RELATIONSHIP = "evident-by"
    CONTAINS_RELATIONSHIP = "contains"
    DEPENDENCY_OF_RELATIONSHIP = "dependency-of"
    DESCRIBED_BY_RELATIONSHIP = "described-by"


class Relationship(BaseModel):
    from_: str
    to_: str
    type: RelationshipType
    model_config = ConfigDict(frozen=True)



import phpserialize

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_pecl
from labels.parsers.cataloger.utils import get_enriched_location


def parse_pecl_serialized(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    unserialized_data = phpserialize.loads(reader.read_closer.read().encode(), decode_strings=True)

    packages = _collect_packages(unserialized_data, reader)

    return packages, []


def _collect_packages(
    unserialized_data: dict[str, object], reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []

    name = str(unserialized_data.get("name", "")) or None

    version_data = unserialized_data.get("version")
    version = (
        (str(version_data.get("release", "")) or None) if isinstance(version_data, dict) else None
    )

    new_location = get_enriched_location(
        reader.location,
        package_manager=PackageManager.COMPOSER,
    )

    package = new_package_from_pecl(name=name, version=version, location=new_location)
    if package:
        packages.append(package)

    return packages



import abc
from collections.abc import Generator
from typing import Protocol, TextIO, runtime_checkable

from labels.model.file import Location


class FileReader(Protocol):  # pragma: no cover
    def __call__(self, filename: str, *, encoding: str, mode: str) -> TextIO: ...


@runtime_checkable
class ContentResolverProtocol(Protocol):  # pragma: no cover
    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str = "r",
    ) -> TextIO | None: ...


class ContentResolver(metaclass=abc.ABCMeta):  # pragma: no cover
    @classmethod
    def __subclasshook__(cls: type["ContentResolver"], subclass: object) -> bool:
        return bool(isinstance(subclass, ContentResolverProtocol))

    @abc.abstractmethod
    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str = "r",
    ) -> TextIO | None:
        """Resolve file context."""
        raise NotImplementedError


@runtime_checkable
class PathResolverProtocol(Protocol):  # pragma: no cover
    def has_path(self, path: str) -> bool: ...

    def files_by_path(self, *paths: str) -> list[Location]: ...

    def files_by_glob(self, *patters: str) -> list[Location]: ...

    def relative_file_path(self, _: Location, path: str) -> Location | None: ...

    def walk_file(self) -> Generator[str, None, str]: ...


class PathResolver(metaclass=abc.ABCMeta):  # pragma: no cover
    @classmethod
    def __subclasshook__(cls: type["PathResolver"], subclass: object) -> bool:
        return bool(isinstance(subclass, PathResolverProtocol))

    @abc.abstractmethod
    def has_path(self, path: str) -> bool:
        """Resolve file context."""
        raise NotImplementedError

    @abc.abstractmethod
    def files_by_path(self, *paths: str) -> list[Location]:
        """Resolve file context."""
        raise NotImplementedError

    @abc.abstractmethod
    def files_by_glob(self, *patters: str) -> list[Location]:
        """Resolve file context."""
        raise NotImplementedError

    @abc.abstractmethod
    def relative_file_path(self, _: Location, path: str) -> Location | None:
        """Resolve file context."""
        raise NotImplementedError

    @abc.abstractmethod
    def walk_file(self) -> Generator[str, None, None]:
        raise NotImplementedError


class Resolver(ContentResolver, PathResolver, metaclass=abc.ABCMeta):  # pragma: no cover
    pass



import re
from typing import NamedTuple

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProject
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location

QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r"[a-zA-Z0-9._-]+"

CONFIG_TO_DEV_STATUS = {
    "testRuntimeOnly": True,
    "testCompileOnly": True,
    "testImplementation": True,
    "compileOnly": True,
    "testCompile": True,
    "androidTestImplementation": True,
    "compile": False,
    "runtime": False,
    "implementation": False,
    "api": False,
    "runtimeOnly": False,
}


class GradleKtsDependency(NamedTuple):
    group: str
    name: str
    version: str
    config: str
    line: int | None = None


def parse_gradle_lockfile_kts(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    gradle_dependencies = _parse_dependencies(reader.read_closer.read())

    packages = _collect_packages(gradle_dependencies, reader.location)

    return packages, []


def _parse_dependencies(file_content: str) -> list[GradleKtsDependency]:
    dependencies: list[GradleKtsDependency] = []
    is_block_comment = False

    for line_no, raw_line in enumerate(file_content.splitlines(), start=1):
        line = raw_line.strip()
        is_block_comment = _update_block_comment_status(line, is_block_comment=is_block_comment)

        if is_block_comment or _is_comment(line):
            continue

        dependency = _extract_dependency(line, line_no)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _update_block_comment_status(line: str, *, is_block_comment: bool) -> bool:
    if "/*" in line:
        return True
    if "*/" in line:
        return False
    return is_block_comment


def _is_comment(line: str) -> bool:
    return (
        line.strip().startswith("//")
        or line.strip().startswith("/*")
        or line.strip().endswith("*/")
    )


def _extract_dependency(line: str, line_no: int) -> GradleKtsDependency | None:
    regex = _build_regex_with_configs()

    if match := regex.match(line):
        version = match.group("version")
        config = match.group("config_name")

        return GradleKtsDependency(
            group=match.group("group"),
            name=match.group("name"),
            version=version,
            config=config,
            line=line_no,
        )

    return None


def _build_regex_with_configs() -> re.Pattern[str]:
    configs = {
        "runtimeOnly",
        "api",
        "compile",
        "compileOnly",
        "implementation",
        "testRuntimeOnly",
        "testCompileOnly",
        "testImplementation",
        "runtime",
        "androidTestImplementation",
    }

    config_pattern = "|".join(configs)

    return re.compile(
        rf"(?P<config_name>{config_pattern})\({QUOTE}"
        rf"(?P<group>{TEXT}):(?P<name>{TEXT}):"
        rf"(?P<version>{TEXT})"
        rf"(?::(?P<classifier>{TEXT}))?"
        rf"{QUOTE}\)",
    )


def _collect_packages(
    dependencies: list[GradleKtsDependency],
    reader_location: Location,
) -> list[Package]:
    packages: list[Package] = []

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        is_dev = CONFIG_TO_DEV_STATUS.get(dependency.config)

        new_location = get_enriched_location(
            reader_location,
            line=dependency.line,
            is_dev=is_dev,
            package_manager=PackageManager.GRADLE,
        )

        ecosystem_data = JavaArchive(
            pom_project=JavaPomProject(
                group_id=dependency.group,
                name=name,
                artifact_id=name,
                version=version,
            ),
        )

        package_spec = JavaPackageSpec(
            simple_name=name,
            composed_name=f"{dependency.group}:{name}",
            version=version,
            location=new_location,
            ecosystem_data=ecosystem_data,
        )

        package = new_java_package(package_spec)
        if package:  # pragma: no branch
            packages.append(package)

    return packages



import re
from collections.abc import Iterator
from typing import Final, NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.golang.package_builder import new_go_package
from labels.parsers.cataloger.utils import get_enriched_location

GO_DIRECTIVE: re.Pattern[str] = re.compile(
    r"(?P<directive>require|replace) \(",
)
GO_MOD_DEP: re.Pattern[str] = re.compile(
    r"^\s+(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_REPLACE: re.Pattern[str] = re.compile(
    r"^\s+(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REP_DEP: re.Pattern[str] = re.compile(
    r"replace\s(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REQ_MOD_DEP: re.Pattern[str] = re.compile(
    r"require\s(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_VERSION: re.Pattern[str] = re.compile(
    r"\ngo (?P<major>\d)\.(?P<minor>\d+)(\.\d+)?\n",
)

MINIMUM_SUPPORTED_GO_VERSION: Final[tuple[int, int]] = (1, 17)


class ParseState(NamedTuple):
    replace_list: list[tuple[re.Match[str], int]]
    req_dict: dict[str, Package]


def parse_go_mod(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    file_content = reader.read_closer.read()
    if _is_go_version_supported(file_content):
        packages = list(_collect_packages(file_content, reader.location))

    return packages, []


def _handle_directive_block_line(
    line: str, line_number: int, directive: str, state: ParseState, location: Location
) -> str:
    if directive == "replace":
        if replace := GO_REPLACE.search(line):
            state.replace_list.append((replace, line_number))
            return directive
        return ""
    if matched := GO_MOD_DEP.search(line):
        _add_required_package(matched, state.req_dict, line_number, location)
        return directive
    return ""


def _collect_packages(
    file_content: str,
    location: Location,
) -> Iterator[Package]:
    directive: str = ""
    state = ParseState(replace_list=[], req_dict={})

    for line_number, line in enumerate(file_content.splitlines(), 1):
        if matched := GO_REQ_MOD_DEP.search(line):
            _add_required_package(matched, state.req_dict, line_number, location)
            continue
        if replace := GO_REP_DEP.search(line):
            state.replace_list.append((replace, line_number))
            continue
        if not directive:
            if match := GO_DIRECTIVE.match(line):
                directive = match.group("directive")
            continue
        directive = _handle_directive_block_line(line, line_number, directive, state, location)
    return _apply_replace_directives(state.req_dict, state.replace_list, location)


def _add_required_package(
    matched: re.Match[str],
    req_dict: dict[str, Package],
    line_number: int,
    parent_location: Location,
) -> None:
    product: str = matched.group("product")
    version: str = matched.group("version")

    new_location = get_enriched_location(
        parent_location,
        line=line_number,
        is_transitive=False,
        package_manager=PackageManager.GO_MOD,
    )

    package = new_go_package(product, version, new_location)
    if package:  # pragma: no branch
        req_dict[product] = package


def _apply_replace_directives(
    req_dict: dict[str, Package],
    replace_list: list[tuple[re.Match[str], int]],
    parent_location: Location,
) -> Iterator[Package]:
    for matched, line_number in replace_list:
        match_dict = matched.groupdict()
        old_pkg, old_version = match_dict["old_prod"], match_dict["old_ver"]
        repl_pkg, version = match_dict["new_prod"], match_dict["new_ver"]

        if old_pkg not in req_dict:
            continue

        if old_version and not version:
            version = req_dict[old_pkg].version

        if not version or (old_version and req_dict[old_pkg].version != old_version):
            continue

        new_location = get_enriched_location(
            parent_location,
            line=line_number,
            is_transitive=False,
            package_manager=PackageManager.GO_MOD,
        )

        package = new_go_package(repl_pkg, version, new_location)
        if package:  # pragma: no branch
            req_dict[old_pkg] = package

    return iter(req_dict.values())


def _is_go_version_supported(content: str) -> bool:
    go_version = GO_VERSION.search(content)
    if not go_version:
        return False

    major = int(go_version.group("major"))
    minor = int(go_version.group("minor"))

    return (major, minor) >= MINIMUM_SUPPORTED_GO_VERSION



import tempfile
import zipfile
from pathlib import Path
from typing import TextIO

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProperties
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.utils.zip import safe_extract


def parse_apk(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    with tempfile.TemporaryDirectory() as tmp_dir:
        file_paths = _extract_apk_to_tempdir(reader, tmp_dir)
        if not file_paths:
            return [], []

        packages = _collect_packages(file_paths, reader.location)

    return packages, []


def _extract_apk_to_tempdir(reader: LocationReadCloser, output_folder: str) -> list[Path]:
    files_paths: list[Path] = []
    read_closer: TextIO = reader.read_closer
    filename: str = read_closer.name
    try:
        with zipfile.ZipFile(filename, "r") as apk_file:
            safe_extract(apk_file, output_folder)
    except zipfile.BadZipFile:
        return files_paths

    meta_dir_path = Path(output_folder) / "META-INF"
    if meta_dir_path.exists():
        files_paths = [
            file_path
            for file_path in meta_dir_path.iterdir()
            if file_path.name.endswith(".version")
        ]

    return files_paths


def _collect_packages(file_paths: list[Path], location: Location) -> list[Package]:
    packages: list[Package] = []

    for file_path in file_paths:
        result = _read_version_and_ids(file_path)
        if not result:
            continue

        group_id, artifact_id, version = result

        java_archive = JavaArchive(
            pom_properties=JavaPomProperties(
                group_id=group_id,
                artifact_id=artifact_id,
                version=version,
            ),
        )

        new_location = get_enriched_location(location)

        package_spec = JavaPackageSpec(
            simple_name=artifact_id,
            composed_name=f"{group_id}:{artifact_id}",
            version=version,
            location=new_location,
            ecosystem_data=java_archive,
        )

        package = new_java_package(package_spec)
        if package:
            packages.append(package)

    return packages


def _read_version_and_ids(file_path: Path) -> tuple[str, str, str] | None:
    with file_path.open(encoding="utf-8") as f:
        version = f.read().strip()

    parts = file_path.name.replace(".version", "").split("_", 1)

    if len(parts) < 2:
        return None

    group_id, artifact_id = parts

    if not group_id or not artifact_id:
        return None

    return group_id, artifact_id, version



import re

from labels.model.package import Ecosystem


def format_exception(exc: str) -> str:
    return re.sub(r"^\s*For further information.*(?:\n|$)", "", exc, flags=re.MULTILINE)


def normalize_python_package_name(name: str) -> str:
    return re.sub(r"[-_.]+", "-", name)


def normalize_name(name: str, ecosystem: Ecosystem) -> str:
    if ecosystem == Ecosystem.NUGET:
        return name.lower()

    if ecosystem == Ecosystem.PYPI:
        return normalize_python_package_name(name.lower())

    return name



from typing import NamedTuple

from labels.model.file import DependencyType, Location
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.parsers.cataloger.javascript.common import (
    is_version_compatible,
)
from labels.parsers.cataloger.javascript.package_builder import new_npm_package_from_lock
from labels.parsers.cataloger.utils import get_enriched_location, parse_integrity_to_health_metadata


class PackageV2Info(NamedTuple):
    path: str
    dependencies: dict[str, str]
    dev_dependencies: dict[str, str]
    peer_dependencies: dict[str, str]
    optional_dependencies: dict[str, str]


class _RelationshipContext(NamedTuple):
    package: Package
    package_location: Location
    pkg_info: PackageV2Info
    packages_by_name: dict[str, list[Package]]
    lockfile_path: str
    package_map: dict[tuple[str, int], PackageV2Info]


class LocationLookup(NamedTuple):
    dep_locations: list[Location]
    package_map: dict[tuple[str, int], PackageV2Info]


def parse_package_lock_v2(
    location: Location,
    file_content: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return [], []

    packages = []
    root_package = _create_root_package(location, packages_dict)
    if root_package:
        packages.append(root_package)

    direct_dependencies = _get_direct_dependencies_from_root(packages_dict)
    packages.extend(_collect_packages(location, packages_dict, direct_dependencies))
    package_map_v2 = _build_package_map_v2(packages_dict)
    relationships = _build_relationships_v2(packages, package_map_v2, location.path())

    return packages, relationships


def _collect_packages(
    location: Location,
    packages_dict: IndexedDict[str, ParsedValue],
    direct_dependencies: list[str],
) -> list[Package]:
    packages: list[Package] = []

    for dependency_key, package_value in packages_dict.items():
        if dependency_key == "":
            continue

        if not isinstance(package_value, IndexedDict):
            continue

        name = _get_name(dependency_key, package_value) or _get_name_from_path(dependency_key)

        is_transitive = name not in direct_dependencies
        is_dev = package_value.get("dev") is True
        new_location = get_enriched_location(
            location,
            line=package_value.position.start.line,
            is_dev=is_dev,
            is_transitive=is_transitive,
            package_manager=PackageManager.NPM,
        )

        integrity_str = package_value.get("integrity")
        health_metadata = parse_integrity_to_health_metadata(
            str(integrity_str) if integrity_str else None
        )

        package = new_npm_package_from_lock(
            location=new_location,
            name=name,
            value=package_value,
            lockfile_version=2,
            health_metadata=health_metadata,
        )
        if package:
            packages.append(package)

    return packages


def _extract_dict_from_indexed(value: ParsedValue | None) -> dict[str, str]:
    if isinstance(value, IndexedDict):
        return {k: str(v) for k, v in value.data.items()}
    return {}


def _build_package_map_v2(
    packages_dict: IndexedDict[str, ParsedValue],
) -> dict[tuple[str, int], PackageV2Info]:
    package_map: dict[tuple[str, int], PackageV2Info] = {}

    for package_path, package_value in packages_dict.items():
        if not isinstance(package_value, IndexedDict):
            continue

        name = _get_name(package_path, package_value)
        if not name:
            continue

        line = package_value.position.start.line
        package_map[(name, line)] = PackageV2Info(
            path=package_path,
            dependencies=_extract_dict_from_indexed(package_value.get("dependencies")),
            dev_dependencies=_extract_dict_from_indexed(package_value.get("devDependencies")),
            peer_dependencies=_extract_dict_from_indexed(package_value.get("peerDependencies")),
            optional_dependencies=_extract_dict_from_indexed(
                package_value.get("optionalDependencies")
            ),
        )
    return package_map


def _create_root_package(
    location: Location,
    packages_dict: IndexedDict[str, ParsedValue],
) -> Package | None:
    root_entry = packages_dict.get("")
    if not isinstance(root_entry, IndexedDict):
        return None

    root_name = root_entry.get("name")
    if not isinstance(root_name, str):
        return None

    root_location = get_enriched_location(
        base=location,
        line=root_entry.position.start.line,
        is_transitive=False,
        is_dev=False,
        package_manager=PackageManager.NPM,
    )
    update_root: dict[str, DependencyType] = {"dependency_type": DependencyType.ROOT}
    root_location = root_location.model_copy(deep=True, update=update_root)
    return new_npm_package_from_lock(
        location=root_location,
        name=root_name,
        value=root_entry,
        lockfile_version=2,
    )


def _get_direct_dependencies_from_root(
    packages_dict: IndexedDict[str, ParsedValue],
) -> list[str]:
    root_entry = packages_dict.get("")
    if not isinstance(root_entry, IndexedDict):
        return []

    result: list[str] = []

    deps_candidate = root_entry.get("dependencies")
    if isinstance(deps_candidate, IndexedDict):
        result.extend(deps_candidate.keys())

    dev_deps_candidate = root_entry.get("devDependencies")
    if isinstance(dev_deps_candidate, IndexedDict):
        result.extend(dev_deps_candidate.keys())

    return result


def _get_name(dependency_key: str, package_value: IndexedDict[str, ParsedValue]) -> str:
    name = dependency_key
    if not name:
        if "name" not in package_value:
            return _get_name_from_path(dependency_key)
        name = str(package_value["name"])

    # Handle alias name
    if "name" in package_value and package_value["name"] != dependency_key:
        name = str(package_value["name"])

    return _get_name_from_path(name)


def _get_name_from_path(name: str) -> str:
    return name.split("node_modules/")[-1]


def _build_candidate_paths(parent_path: str, name: str) -> list[str]:
    paths = [f"{parent_path}/node_modules/{name}"]
    parts = parent_path.split("/node_modules/")
    for i in range(len(parts) - 1, 0, -1):
        ancestor_path = "/node_modules/".join(parts[:i])
        paths.append(f"{ancestor_path}/node_modules/{name}")
    paths.append(f"node_modules/{name}")
    return paths


def _find_matching_location(
    candidate_paths: list[str],
    dependency_name: str,
    lookup: LocationLookup,
) -> Location | None:
    for candidate_path in candidate_paths:
        for dep_location in lookup.dep_locations:
            line = dep_location.coordinates.line if dep_location.coordinates else None
            if line is None:
                continue
            pkg_info = lookup.package_map.get((dependency_name, line))
            if pkg_info and pkg_info.path == candidate_path:
                return dep_location
    return None


def _resolve_npm_location_v2(
    parent_path: str,
    dependency_name: str,
    lookup: LocationLookup,
    alias_name: str | None = None,
) -> list[Location]:
    if not lookup.dep_locations:
        return []

    candidate_paths = _build_candidate_paths(parent_path, dependency_name)
    if alias_name and alias_name != dependency_name:
        candidate_paths.extend(_build_candidate_paths(parent_path, alias_name))

    location = _find_matching_location(candidate_paths, dependency_name, lookup)
    return [location] if location else []


def _process_dependencies_for_relationships(
    dependencies_dict: dict[str, str],
    context: _RelationshipContext,
    relationships: list[Relationship],
) -> None:
    for original_dep_name, raw_required_version in dependencies_dict.items():
        real_name, normalized_required_version = _parse_alias_requirement(raw_required_version)
        lookup_name = real_name or original_dep_name

        for dep_pkg in context.packages_by_name.get(lookup_name, []):
            if not is_version_compatible(dep_pkg.version, normalized_required_version):
                continue
            all_locs = [loc for loc in dep_pkg.locations if loc.path() == context.lockfile_path]
            lookup = LocationLookup(dep_locations=all_locs, package_map=context.package_map)
            dep_locations = _resolve_npm_location_v2(
                context.pkg_info.path,
                lookup_name,
                lookup,
                original_dep_name if real_name else None,
            )
            for dep_location in dep_locations:
                from_id = f"{dep_pkg.id_}@{dep_location.location_id()}"
                to_id = f"{context.package.id_}@{context.package_location.location_id()}"
                relationships.append(
                    Relationship(
                        from_=from_id,
                        to_=to_id,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                )


def _build_relationships_v2(
    packages: list[Package],
    package_map: dict[tuple[str, int], PackageV2Info],
    lockfile_path: str,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_name: dict[str, list[Package]] = {}
    for pkg in packages:
        if any((loc.path() == lockfile_path) for loc in pkg.locations):
            packages_by_name.setdefault(pkg.name, []).append(pkg)

    for package in packages:
        for package_location in [loc for loc in package.locations if loc.path() == lockfile_path]:
            line = package_location.coordinates.line if package_location.coordinates else None
            if line is None:
                continue

            pkg_info = package_map.get((package.name, line))
            if not pkg_info:
                continue

            context = _RelationshipContext(
                package=package,
                package_location=package_location,
                pkg_info=pkg_info,
                packages_by_name=packages_by_name,
                lockfile_path=lockfile_path,
                package_map=package_map,
            )

            for dependencies_dict in [
                pkg_info.dependencies,
                pkg_info.dev_dependencies,
                pkg_info.peer_dependencies,
                pkg_info.optional_dependencies,
            ]:
                _process_dependencies_for_relationships(
                    dependencies_dict,
                    context,
                    relationships,
                )

    return relationships


def _parse_alias_requirement(required_version: str) -> tuple[str | None, str]:
    """Normalize npm alias requirements.

    For entries like "npm:strip-ansi@^6.0.1" return ("strip-ansi", "^6.0.1").
    For non-alias requirements, return (None, required_version).
    """
    version_str = required_version.strip()
    prefix = "npm:"
    if not version_str.startswith(prefix):
        return None, version_str

    tail = version_str[len(prefix) :]
    if "@" in tail:
        real_name, parsed_version = tail.split("@", 1)
        return real_name, parsed_version.strip()
    return tail, "*"



import logging
import tempfile
from collections.abc import Generator
from contextlib import contextmanager, suppress
from pathlib import Path
from zipfile import ZipFile, ZipInfo

from labels.model.ecosystem_data.java import JavaArchive
from labels.model.file import Location
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.parsers.cataloger.java.parse_archive.archive_file_info import (
    ArchiveFileInfo,
    parse_file_info,
)
from labels.parsers.cataloger.java.parse_archive.manifest_parser import ManifestParser
from labels.parsers.cataloger.java.parse_archive.maven_handler import (
    JavaPackageIdentity,
    MavenHandler,
)
from labels.parsers.cataloger.java.utils.group_id_resolver import group_id_from_java_metadata
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.utils.zip import new_zip_file_manifest, new_zip_glob_match, traverse_files_in_zip

LOGGER = logging.getLogger(__name__)

NESTED_FILES_EXTENSIONS = (
    "*.jar",
    "*.war",
    "*.ear",
    "*.par",
    "*.sar",
    "*.nar",
    "*.jpi",
    "*.hpi",
    "*.kar",
    "*.lpkg",
)


class ArchiveParser:
    def __init__(
        self,
        *,
        file_manifest: list[ZipInfo],
        location: Location,
        archive_path: str,
        file_info: ArchiveFileInfo,
        detect_nested: bool,
    ) -> None:
        self.file_manifest = file_manifest
        self.location = location
        self.archive_path = archive_path
        self.file_info = file_info
        self.detect_nested = detect_nested

        self.manifest_parser = ManifestParser(
            file_manifest=file_manifest,
            archive_path=archive_path,
            location=location,
        )
        self.maven_handler = MavenHandler(
            file_manifest=file_manifest,
            archive_path=archive_path,
            location=location,
            file_info=file_info,
        )
        self.nested_handler = NestedArchiveHandler(
            file_manifest=file_manifest,
            archive_path=archive_path,
            location=location,
        )

    def parse(self) -> tuple[list[Package], list[Relationship]]:
        packages: list[Package] = []
        relationships: list[Relationship] = []

        parent_package = self._discover_main_package()

        if parent_package:
            auxiliary_packages = self.maven_handler.discover_auxiliary_packages(parent_package)
            packages.append(parent_package)

            for auxiliary_package in auxiliary_packages:
                packages.append(auxiliary_package)
                relationships.append(
                    Relationship(
                        from_=auxiliary_package.id_,
                        to_=parent_package.id_,
                        type=RelationshipType.CONTAINS_RELATIONSHIP,
                    )
                )

        if self.detect_nested:
            nested_packages, nested_relationships = self.nested_handler.discover_packages()
            packages.extend(nested_packages)
            relationships.extend(nested_relationships)

        return packages, relationships

    def _discover_main_package(self) -> Package | None:
        manifest = self.manifest_parser.parse()
        if not manifest:
            return None

        package_identity = self.maven_handler.extract_identity(manifest)
        ecosystem_data = self.maven_handler.build_ecosystem_data(manifest)
        package_spec = self._build_package_spec(package_identity, ecosystem_data)

        return new_java_package(package_spec)

    def _build_package_spec(
        self,
        package_identity: JavaPackageIdentity,
        ecosystem_data: JavaArchive,
    ) -> JavaPackageSpec:
        artifact_id = package_identity.artifact_id
        version = package_identity.version
        group_id = package_identity.group_id

        new_location = get_enriched_location(self.location, package_manager=PackageManager.MAVEN)

        authoritative_group_id = (
            group_id_from_java_metadata(artifact_id, ecosystem_data) or group_id
        )

        return JavaPackageSpec(
            simple_name=artifact_id,
            version=version,
            location=new_location,
            composed_name=f"{authoritative_group_id}:{artifact_id}",
            ecosystem_data=ecosystem_data,
        )


class NestedArchiveHandler:
    def __init__(
        self, *, file_manifest: list[ZipInfo], archive_path: str, location: Location
    ) -> None:
        self.file_manifest = file_manifest
        self.archive_path = archive_path
        self.location = location

    def discover_packages(self) -> tuple[list[Package], list[Relationship]]:
        packages: list[Package] = []
        relationships: list[Relationship] = []

        nested_paths = self._find_nested_archives()
        if not nested_paths:
            return packages, relationships

        for path in nested_paths:
            try:
                nested_packages, nested_relationships = self._process_single_archive(path)
                packages.extend(nested_packages)
                relationships.extend(nested_relationships)
            except (OSError, ValueError, KeyError) as ex:
                LOGGER.debug("Unable to process nested archive %s: %s", path, ex)
                continue

        return packages, relationships

    def _find_nested_archives(self) -> list[str]:
        return new_zip_glob_match(self.file_manifest, NESTED_FILES_EXTENSIONS, case_sensitive=True)

    def _process_single_archive(
        self, archive_path: str
    ) -> tuple[list[Package], list[Relationship]]:
        nested_bytes = self._extract_archive_bytes(archive_path)
        if not nested_bytes:  # pragma: no cover
            return [], []

        with self._create_temp_file(archive_path, nested_bytes) as temp_path:
            nested_location = Location(
                coordinates=self.location.coordinates,
                access_path=f"{self.location.access_path}:{archive_path}",
            )

            nested_parser = ArchiveParser(
                file_manifest=new_zip_file_manifest(temp_path),
                location=nested_location,
                archive_path=temp_path,
                file_info=parse_file_info(archive_path),
                detect_nested=True,
            )

            return nested_parser.parse()

    def _extract_archive_bytes(self, file_path: str) -> bytes | None:
        result = None

        def visitor(file: ZipInfo) -> None:
            nonlocal result
            with ZipFile(self.archive_path, "r") as zip_reader, zip_reader.open(file) as file_data:
                result = file_data.read()

        traverse_files_in_zip(self.archive_path, visitor, file_path)
        return result

    @contextmanager
    def _create_temp_file(self, archive_path: str, content: bytes) -> Generator[str]:
        suffix = f"_{Path(archive_path).name}"
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as tmp:
            tmp.write(content)
            temp_path = tmp.name

        try:
            yield temp_path
        finally:
            with suppress(OSError):
                Path(temp_path).unlink()



from pydantic import BaseModel

from labels.model.constraints import NonEmptyStr


class Digest(BaseModel):
    algorithm: NonEmptyStr | None = None
    value: NonEmptyStr | None = None


class Artifact(BaseModel):
    integrity: Digest | None = None


class HealthMetadata(BaseModel):
    artifact: Artifact | None = None



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning


def new_package_from_composer(
    *,
    name: str | None,
    version: str | None,
    location: Location,
) -> Package | None:
    if not name or not version:
        return None

    p_url = _package_url_from_composer(name, version)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.PHP,
            ecosystem=Ecosystem.PACKAGIST,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def new_package_from_pecl(
    *,
    name: str | None,
    version: str | None,
    location: Location,
) -> Package | None:
    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.PHP,
            ecosystem=Ecosystem.PACKAGIST,
            p_url=PackageURL(type="pecl", name=name, version=version).to_string(),
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _package_url_from_composer(name: str, version: str) -> str:
    fields = name.split("/")

    vendor = ""
    if len(fields) == 1:
        name = fields[0]
    else:
        vendor = fields[0]
        name = "-".join(fields[1:])

    return PackageURL(type="composer", namespace=vendor, name=name, version=version).to_string()



import re

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.toml import parse_toml_with_tree_sitter


def parse_pyproject_toml(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_toml_with_tree_sitter(reader.read_closer.read())

    packages = _collect_packages(file_content, reader)

    return packages, []


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    return [
        *_parse_poetry_dependencies(file_content, reader),
        *_parse_uv_dependencies(file_content, reader),
    ]


def _parse_poetry_dependencies(
    file_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages: list[Package] = []
    tool = file_content.get("tool")
    if not isinstance(tool, IndexedDict):
        return packages

    poetry = tool.get("poetry")
    if not isinstance(poetry, IndexedDict):
        return packages

    deps = poetry.get("dependencies")
    if isinstance(deps, IndexedDict):
        packages.extend(_get_poetry_packages(reader, deps))

    group = poetry.get("group")
    if isinstance(group, IndexedDict):
        dev = group.get("dev")
        if isinstance(dev, IndexedDict):
            dev_deps = dev.get("dependencies")
            if isinstance(dev_deps, IndexedDict):
                packages.extend(_get_poetry_packages(reader, dev_deps, is_dev=True))

    dev_dependencies = poetry.get("dev-dependencies")
    if isinstance(dev_dependencies, IndexedDict):
        packages.extend(_get_poetry_packages(reader, dev_dependencies, is_dev=True))

    return packages


def _get_poetry_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool = False,
) -> list[Package]:
    packages: list[Package] = []

    items = dependencies.items()

    for name, value in items:
        version = _get_version(value)

        new_location = get_enriched_location(
            reader.location,
            line=dependencies.get_key_position(name).start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.PIP,
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            packages.append(package)

    return packages


def _get_version(value: ParsedValue) -> str | None:
    if isinstance(value, str):
        return value
    if not isinstance(value, IndexedDict):
        return None
    return str(value.get("version", ""))


def _parse_uv_dependencies(
    content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages: list[Package] = []
    project = content.get("project")
    if not isinstance(project, IndexedDict):
        return packages

    uv_deps = project.get("dependencies")
    if isinstance(uv_deps, IndexedList):
        packages.extend(_get_uv_packages(reader, uv_deps))

    optional_deps = project.get("optional-dependencies")
    if isinstance(optional_deps, IndexedDict):
        uv_dev_deps = optional_deps.get("dev")
        if isinstance(uv_dev_deps, IndexedList):
            packages.extend(_get_uv_packages(reader, uv_dev_deps, is_dev=True))

    dependency_groups = content.get("dependency-groups")
    if isinstance(dependency_groups, IndexedDict):
        dev_group = dependency_groups.get("dev")
        if isinstance(dev_group, IndexedList):
            packages.extend(_get_uv_packages(reader, dev_group, is_dev=True))

    return packages


def _get_uv_packages(
    reader: LocationReadCloser,
    dependencies: IndexedList[ParsedValue],
    *,
    is_dev: bool = False,
) -> list[Package]:
    packages: list[Package] = []
    dep_pattern = re.compile(r"^([A-Za-z0-9_.\-]+)\s*([<>=!~]+.*)?$")

    for index, dep_string in enumerate(dependencies):
        if not isinstance(dep_string, str):
            continue

        match = dep_pattern.match(dep_string.strip())
        if match:
            name = match.group(1)
            version = str(match.group(2)).strip() if match.group(2) else "*"
        else:
            name = dep_string.strip()
            version = "*"

        new_location = get_enriched_location(
            reader.location,
            line=dependencies.get_position(index).start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.PIP,
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            packages.append(package)

    return packages



from typing import NamedTuple, TextIO

from labels.model.ecosystem_data.python import WheelEggEcosystemData
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package


class WheelEggPackageData(NamedTuple):
    name: str
    version: str
    ecosystem_data: WheelEggEcosystemData | None = None


def parse_wheel_or_egg(
    resolver: Resolver,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_data = _assemble_wheel_egg_package_data(resolver, reader.location)
    if not package_data:
        return [], []

    packages = _collect_packages(package_data, reader)

    return packages, []


def _collect_packages(
    package_data: WheelEggPackageData, reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []
    package = new_python_package(
        name=package_data.name,
        version=package_data.version,
        location=reader.location,
        ecosystem_data=package_data.ecosystem_data,
    )
    if package:  # pragma: no branch
        packages.append(package)

    return packages


def _assemble_wheel_egg_package_data(
    resolver: Resolver,
    metadata_location: Location,
) -> WheelEggPackageData | None:
    metadata_content = resolver.file_contents_by_location(metadata_location)

    if not metadata_content or not metadata_location.coordinates:
        return None

    package_data = _parse_wheel_egg_package_data(metadata_content)
    if not package_data:
        return None

    return package_data


def _parse_wheel_egg_package_data(reader: TextIO) -> WheelEggPackageData | None:
    metadata_dict = _parse_metadata(reader.read())

    name = metadata_dict.get("Name")
    version = metadata_dict.get("Version")

    if not isinstance(name, str) or not isinstance(version, str):
        return None

    dependencies = _extract_dependencies(metadata_dict)

    return WheelEggPackageData(
        name=name,
        version=version,
        ecosystem_data=WheelEggEcosystemData(dependencies=dependencies),
    )


def _extract_dependencies(metadata_dict: dict[str, str | list[str]]) -> list[str]:
    dependencies: list[str] = []

    requires_dist = metadata_dict.get("Requires-Dist")
    provides_extra = metadata_dict.get("Provides-Extra")

    if not isinstance(provides_extra, list):
        provides_extra = None

    if requires_dist:
        dependencies = _required_dependencies(requires_dist, provides_extra)

    return dependencies


def _parse_metadata(data: str) -> dict[str, str | list[str]]:
    parsed_data: dict[str, str | list[str]] = {}
    lines = _split_lines(data)
    multi_line_key: str | None = None

    for line in lines:
        if not line:
            break
        if multi_line_key and line.startswith((" ", "\t")):
            multi_line_key = _handle_multiline_values(line, multi_line_key, parsed_data)
        else:
            multi_line_key = _process_line(line, parsed_data)
    return parsed_data


def _split_lines(data: str) -> list[str]:
    return data.strip().split("\n")


def _process_line(line: str, parsed_data: dict[str, str | list[str]]) -> str:
    multi_line_key = None
    if ": " in line:
        key, value = line.split(": ", 1)
        _update_parsed_data(key, value, parsed_data)
        if key in ["Description", "Classifier"]:
            multi_line_key = key
    return multi_line_key or ""


def _handle_multiline_values(
    line: str,
    multi_line_key: str,
    parsed_data: dict[str, str | list[str]],
) -> str:
    current_value = parsed_data.get(multi_line_key)
    if multi_line_key and isinstance(current_value, str):
        parsed_data[multi_line_key] = current_value + "\n" + line.strip()
    return multi_line_key


def _update_parsed_data(key: str, value: str, parsed_data: dict[str, str | list[str]]) -> None:
    if key in parsed_data:
        existing_value = parsed_data[key]
        if isinstance(existing_value, list):
            existing_value.append(value)
        else:
            parsed_data[key] = [existing_value, value]
    else:
        parsed_data[key] = value


def _required_dependencies(
    requires_dis: list[str] | str,
    provides_extra: list[str] | None = None,
) -> list[str]:
    if isinstance(requires_dis, str):
        requires_dis = [requires_dis]
    result: list[str] = []
    provides_extra = provides_extra or []
    for item in requires_dis:
        parts = item.split(";")
        if any(x in parts[-1] for x in provides_extra):
            continue
        result.append(parts[0].strip())
    return result



import glob
import os
import re
from collections.abc import Generator
from fnmatch import translate
from pathlib import Path
from typing import TextIO, cast

from labels.model.file import Coordinates, Location
from labels.model.resolver import ContentResolver, FileReader, PathResolver, Resolver

REPLACE_SET = {
    "[": "[[]",
    "]": "[]]",
}


def _normalize_rules(rules: tuple[str, ...]) -> tuple[str, ...]:
    normalized_rules: list[str] = []
    for rule in rules:
        new_rule = rule
        if new_rule.startswith("glob(") and new_rule.endswith(")"):
            new_rule = new_rule[5:-1]
        elif "[" in new_rule:
            new_rule = "".join(REPLACE_SET.get(char, char) for char in new_rule)
        if new_rule == ".":
            new_rule = "**"
        if new_rule.endswith("/"):
            new_rule += "**"
        if new_rule.startswith("**/"):
            normalized_rules.append(new_rule[3:])
        normalized_rules.append(new_rule)
    return tuple(normalized_rules)


class DirectoryPathResolver(PathResolver):
    root: str
    include: tuple[str, ...]
    exclude: tuple[str, ...]

    def __init__(
        self,
        root: str,
        include: tuple[str, ...] = (".",),
        exclude: tuple[str, ...] = (),
    ) -> None:
        self.root = os.path.realpath(root)
        self.include = include
        self.exclude = exclude

    def has_path(self, path: str) -> bool:
        return (Path(self.root) / path.lstrip("/")).exists()

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []
        for path in paths:
            relative_path = path.replace(self.root, "").lstrip("/")
            full_path = os.path.join(self.root, relative_path)  # noqa: PTH118
            if Path(full_path).exists():
                locations.append(
                    Location(
                        coordinates=Coordinates(real_path=full_path, file_system_id=""),
                        access_path=relative_path,
                    ),
                )
        return locations

    def files_by_glob(self, *patters: str) -> list[Location]:
        result: list[Location] = []
        for pattern in patters:
            result.extend(
                Location(
                    coordinates=Coordinates(
                        real_path=os.path.join(self.root, item),  # noqa: PTH118
                        file_system_id="",
                    ),
                    access_path=item,
                )
                for item in glob.glob(pattern, root_dir=self.root, recursive=True)  # noqa: PTH207
            )

        return result

    def relative_file_path(self, _: Location, _path: str) -> Location:
        real_path = os.path.realpath(_path)
        return Location(
            coordinates=Coordinates(real_path=real_path, file_system_id="", line=None),
            access_path=real_path.replace(self.root, "").lstrip("/"),
        )

    def walk_file(self) -> Generator[str, None, None]:
        excluded_dirs = ["node_modules", "dist", "__pycache__"]
        exclude_regex = [
            translate(os.path.join(self.root, rule))  # noqa: PTH118
            for rule in _normalize_rules(self.exclude)
        ]
        include_regex = [
            translate(os.path.join(self.root, rule))  # noqa: PTH118
            for rule in _normalize_rules(self.include)
        ]

        for dirpath, _, filenames in os.walk(self.root):
            if any(
                (
                    dirpath.endswith(excluded_dir)
                    or f"{os.path.sep}{excluded_dir}{os.path.sep}" in dirpath
                )
                for excluded_dir in excluded_dirs
            ):
                continue

            for filename in filenames:
                full_path = os.path.join(dirpath, filename)  # noqa: PTH118
                if any(re.match(regex, full_path) for regex in include_regex) and not any(
                    re.match(regex, full_path) for regex in exclude_regex
                ):
                    relative_path = full_path.replace(self.root, "").lstrip("/")
                    yield relative_path


class DirectoryContentResolver(ContentResolver):
    root: str

    def __init__(self, root: str) -> None:
        self.root = os.path.realpath(root)

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str = "r",
    ) -> TextIO | None:
        if not location.coordinates:
            return None

        location_path = Path(location.coordinates.real_path)

        if location_path.exists():
            if function_reader:
                return function_reader(location.coordinates.real_path, encoding="utf-8", mode=mode)

            return cast("TextIO", location_path.open(encoding="utf-8", mode=mode))

        return None


class Directory(Resolver):
    root: str
    include: tuple[str, ...]
    exclude: tuple[str, ...]
    _path_resolver: DirectoryPathResolver
    _content_resolver: DirectoryContentResolver

    def __init__(
        self,
        root: str = "./",
        include: tuple[str, ...] = (".",),
        exclude: tuple[str, ...] = (),
    ) -> None:
        self.root = os.path.realpath(root)
        self.include = include
        self.exclude = exclude
        self._path_resolver = DirectoryPathResolver(root=root, include=include, exclude=exclude)
        self._content_resolver = DirectoryContentResolver(root=root)

    def has_path(self, path: str) -> bool:
        return self._path_resolver.has_path(path)

    def files_by_path(self, *paths: str) -> list[Location]:
        return self._path_resolver.files_by_path(*paths)

    def files_by_glob(self, *patters: str) -> list[Location]:
        return self._path_resolver.files_by_glob(*patters)

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str = "r",
    ) -> TextIO | None:
        return self._content_resolver.file_contents_by_location(
            location,
            function_reader=function_reader,
            mode=mode,
        )

    def relative_file_path(self, loc: Location, path: str) -> Location:
        return self._path_resolver.relative_file_path(loc, path)

    def walk_file(self) -> Generator[str, None, None]:
        yield from self._path_resolver.walk_file()



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.java.parse_android_apk import parse_apk
from labels.parsers.cataloger.java.parse_archive.archive_dispatcher import parse_java_archive
from labels.parsers.cataloger.java.parse_build_sbt import parse_build_sbt
from labels.parsers.cataloger.java.parse_gradle import parse_gradle
from labels.parsers.cataloger.java.parse_gradle_kts import parse_gradle_lockfile_kts
from labels.parsers.cataloger.java.parse_gradle_lockfile import parse_gradle_lockfile
from labels.parsers.cataloger.java.parse_gradle_properties import parse_gradle_properties
from labels.parsers.cataloger.java.parse_pom_xml import parse_pom_xml


def on_next_java(  # mccabe: disable=MC0001
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            patterns = [
                (
                    ("**/*.xml", "*.xml"),
                    parse_pom_xml,
                    "java-parse-pom-xml",
                ),
                (
                    (
                        "**/*.jar",
                        "*.jar",
                        "**/*.war",
                        "*.war",
                        "**/*.ear",
                        "*.ear",
                        "**/*.par",
                        "*.par",
                        "**/*.sar",
                        "*.sar",
                        "**/*.nar",
                        "*.nar",
                        "**/*.jpi",
                        "*.jpi",
                        "**/*.hpi",
                        "*.hpi",
                        "**/*.lpkg",
                        "*.lpkg",
                    ),
                    parse_java_archive,
                    "java-archive-parse",
                ),
                (
                    (
                        "gradle.lockfile*",
                        "**/gradle.lockfile*",
                    ),
                    parse_gradle_lockfile,
                    "java-parse-gradle-lock",
                ),
                (
                    (
                        "**/*.gradle",
                        "/*.gradle",
                        "*.gradle",
                    ),
                    parse_gradle,
                    "java-parse-gradle-lock",
                ),
                (
                    (
                        "**/*.gradle.kts",
                        "*.gradle.kts",
                    ),
                    parse_gradle_lockfile_kts,
                    "java-parse-gradle-kts",
                ),
                (
                    (
                        "**/*.apk",
                        "*.apk",
                    ),
                    parse_apk,
                    "java-parse-apk",
                ),
                (
                    (
                        "**/build.sbt",
                        "build.sbt",
                    ),
                    parse_build_sbt,
                    "java-parse-build-stb",
                ),
                (
                    (
                        "**/gradle-wrapper.properties",
                        "/gradle-wrapper.properties",
                        "gradle-wrapper.properties",
                    ),
                    parse_gradle_properties,
                    "java-parse-gradle-properties",
                ),
            ]
            try:
                for pattern, parser, parser_name in patterns:
                    if any(fnmatch(value, x) for x in pattern):
                        observer.on_next(
                            Request(
                                real_path=value,
                                parser=parser,
                                parser_name=parser_name,
                            ),
                        )
                        break

            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from pydantic import ValidationError

from labels.model.ecosystem_data.python import WheelEggEcosystemData
from labels.model.package import Ecosystem, Package
from labels.model.syft_sbom import SyftArtifact, SyftMetadata
from labels.parsers.builder.utils import get_language, get_locations
from labels.parsers.cataloger.utils import log_malformed_package_warning


def _get_dpkg_entry(metadata: SyftMetadata) -> WheelEggEcosystemData | None:
    if not metadata.requires_dist:
        return None

    dependencies = [
        parts[0].strip()
        for dist in metadata.requires_dist
        if (
            (parts := dist.split(";"))
            and not any(provide in parts[-1] for provide in metadata.provides_extra)
        )
    ]

    return WheelEggEcosystemData(dependencies=dependencies)


def builder(
    artifact: SyftArtifact,
    ecosystem: Ecosystem,
) -> Package | None:
    pkg_language = get_language(artifact.language)
    pkg_locations = get_locations(artifact.locations)

    metadata = artifact.metadata
    pkg_ecosystem_data = _get_dpkg_entry(metadata)

    try:
        package = Package(
            name=artifact.name,
            version=artifact.version,
            p_url=artifact.p_url,
            locations=pkg_locations,
            ecosystem=ecosystem,
            ecosystem_data=pkg_ecosystem_data,
            found_by=artifact.found_by,
            language=pkg_language,
            syft_id=artifact.id_,
        )
    except ValidationError as ex:  # pragma: no cover
        log_malformed_package_warning(pkg_locations[0], ex)
        return None

    return package
